/*
 * Copyright 2013 Michael Osipov
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.michaelo.tomcat.extras.valves;

import java.io.IOException;
import java.io.Writer;
import java.util.Collections;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletResponse;

import org.apache.catalina.connector.Request;
import org.apache.catalina.connector.Response;
import org.apache.catalina.util.RequestUtil;
import org.apache.catalina.util.ServerInfo;
import org.apache.catalina.util.StringManager;
import org.apache.catalina.valves.ErrorReportValve;
import org.apache.commons.lang3.StringUtils;

/**
 * An alternative error report valve with several improvements over the original one.
 * <p>
 * The improvements consist of:
 * <ol>
 * <li>Fix for <a href="https://issues.apache.org/bugzilla/show_bug.cgi?id=54220">bug 54220</a>:
 * ErrorReportValve invoked on non-error responses too. A violation of the servlet specification.</li>
 * <li>Fix for <a href="https://issues.apache.org/bugzilla/show_bug.cgi?id=53071">bug 53071</a>:
 * ErrorReportValve ignores message from throwable. Improves error reporting from throwables.</li>
 * <li>Language style improvements.</li>
 * <li>No repetition of messages.</li>
 * <li>Standard status code reason phrases and descriptions.</li>
 * </ol>
 * </p>
 * This work is based off the original {@link ErrorReportValve} which is licensed under ASLv2.
 *
 * @version $Id: EnhancedErrorReportValve.java 28 2013-03-27 20:56:29Z michael-o $
 */
public class EnhancedErrorReportValve extends ErrorReportValve {

	protected static final StringManager enhancedSm = StringManager
			.getManager(EnhancedErrorReportValve.class.getPackage().getName());

	protected static final SortedSet<Integer> SUPPORTED_STATUS_CODES = Collections
			.unmodifiableSortedSet(new TreeSet<Integer>() {
				private static final long serialVersionUID = 8000886476589541793L;
				{
					add(HttpServletResponse.SC_BAD_REQUEST);
					add(HttpServletResponse.SC_UNAUTHORIZED);
					add(HttpServletResponse.SC_PAYMENT_REQUIRED);
					add(HttpServletResponse.SC_FORBIDDEN);
					add(HttpServletResponse.SC_NOT_FOUND);
					add(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
					add(HttpServletResponse.SC_NOT_ACCEPTABLE);
					add(HttpServletResponse.SC_PROXY_AUTHENTICATION_REQUIRED);
					add(HttpServletResponse.SC_REQUEST_TIMEOUT);
					add(HttpServletResponse.SC_CONFLICT);
					add(HttpServletResponse.SC_GONE);
					add(HttpServletResponse.SC_LENGTH_REQUIRED);
					add(HttpServletResponse.SC_PRECONDITION_FAILED);
					add(HttpServletResponse.SC_REQUEST_ENTITY_TOO_LARGE);
					add(HttpServletResponse.SC_REQUEST_URI_TOO_LONG);
					add(HttpServletResponse.SC_UNSUPPORTED_MEDIA_TYPE);
					add(HttpServletResponse.SC_REQUESTED_RANGE_NOT_SATISFIABLE);
					add(HttpServletResponse.SC_EXPECTATION_FAILED);
					add(422);
					add(423);
					add(424);
					add(426);
					add(428);
					add(429);
					add(431);
					add(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
					add(HttpServletResponse.SC_NOT_IMPLEMENTED);
					add(HttpServletResponse.SC_BAD_GATEWAY);
					add(HttpServletResponse.SC_SERVICE_UNAVAILABLE);
					add(HttpServletResponse.SC_GATEWAY_TIMEOUT);
					add(HttpServletResponse.SC_HTTP_VERSION_NOT_SUPPORTED);
					add(507);
					add(510);
					add(511);
				}
			});

	@Override
	protected void report(Request request, Response response, Throwable throwable) {

		// Do nothing on non-HTTP responses
		int statusCode = response.getStatus();

		// Do nothing on a 1xx, 2xx and 3xx status
		// Do nothing if anything has been written already
		if (statusCode < 400 || response.getContentCount() > 0 || !response.isError())
			return;

		String message;

		if (throwable != null)
			message = RequestUtil.filter(throwable.getMessage());
		else
			message = RequestUtil.filter(response.getMessage());
		if (message == null)
			message = "";

		String reason = null;
		String description = null;

		boolean supportedStatusCode = SUPPORTED_STATUS_CODES.contains(statusCode);

		if (supportedStatusCode) {
			reason = enhancedSm.getString("http." + statusCode + ".reason");
			description = enhancedSm.getString("http." + statusCode + ".desc");
		}

		StringBuffer sb = new StringBuffer();

		sb.append("<html><head><title>");
		sb.append(ServerInfo.getServerInfo()).append(" \u2013 ");
		sb.append(sm.getString("errorReportValve.errorReport"));
		sb.append("</title>");
		sb.append("<style><!--");
		sb.append(org.apache.catalina.util.TomcatCSS.TOMCAT_CSS);
		sb.append("--></style> ");
		sb.append("</head><body>");
		sb.append("<h1>");
		sb.append(enhancedSm.getString("enhancedErrorReportValve.httpStatus"));
		sb.append(' ');
		sb.append(statusCode);
		if (StringUtils.isNotEmpty(reason)) {
			sb.append(" \u2013 ");
			sb.append(reason);
		}
		sb.append("</h1>");
		sb.append("<HR size=\"1\" noshade=\"noshade\">");
		sb.append("<p><b>");
		sb.append(enhancedSm.getString("enhancedErrorReportValve.type"));
		sb.append("</b> ");
		if (throwable != null) {
			sb.append(sm.getString("errorReportValve.exceptionReport"));
		} else {
			sb.append(sm.getString("errorReportValve.statusReport"));
		}
		sb.append("</p>");
		if (StringUtils.isNotEmpty(message)) {
			sb.append("<p><b>");
			sb.append(enhancedSm.getString("enhancedErrorReportValve.message"));
			sb.append("</b> ");
			sb.append(message).append("</p>");
		}
		if (StringUtils.isNotEmpty(description)) {
			sb.append("<p><b>");
			sb.append(enhancedSm.getString("enhancedErrorReportValve.description"));
			sb.append("</b> ");
			sb.append(description);
			sb.append("</p>");
		}

		if (throwable != null) {

			String stackTrace = getPartialServletStackTrace(throwable);
			sb.append("<p><b>");
			sb.append(enhancedSm.getString("enhancedErrorReportValve.exception"));
			sb.append("</b> <pre>");
			sb.append(RequestUtil.filter(stackTrace));
			sb.append("</pre></p>");

			int loops = 0;
			Throwable rootCause = throwable.getCause();
			while (rootCause != null && loops < 10) {
				stackTrace = getPartialServletStackTrace(rootCause);
				sb.append("<p><b>");
				sb.append(enhancedSm.getString("enhancedErrorReportValve.rootCause"));
				sb.append("</b> <pre>");
				sb.append(RequestUtil.filter(stackTrace));
				sb.append("</pre></p>");
				// In case root cause is somehow heavily nested
				rootCause = rootCause.getCause();
				loops++;
			}

			sb.append("<p><b>");
			sb.append(enhancedSm.getString("enhancedErrorReportValve.note"));
			sb.append("</b> ");
			sb.append(sm.getString("errorReportValve.rootCauseInLogs", ServerInfo.getServerInfo()));
			sb.append("</p>");

		}

		sb.append("<HR size=\"1\" noshade=\"noshade\">");
		sb.append("<h3>").append(ServerInfo.getServerInfo()).append("</h3>");
		sb.append("</body></html>");

		try {
			try {
				response.setContentType("text/html");
				response.setCharacterEncoding("utf-8");
			} catch (Throwable t) {
				if (container.getLogger().isDebugEnabled())
					container.getLogger().debug("status.setContentType", t);
			}
			Writer writer = response.getReporter();
			if (writer != null) {
				// If writer is null, it's an indication that the response has
				// been hard committed already, which should never happen
				writer.write(sb.toString());
			}
		} catch (IOException e) {
			; // ignore
		} catch (IllegalStateException e) {
			; // ignore
		}

	}

}
